from enum import Enum
from dataclasses import dataclass
from typing import Optional


class SocialLoginProvider(Enum):
    GOOGLE = "Google"
    GITHUB = "GitHub"
    MICROSOFT = "Microsoft"
    SLACK = "Slack"
    LINKEDIN = "LinkedIn"
    SALESFORCE = "Salesforce"
    XERO = "Xero"
    QUICKBOOKS_ONLINE = "QuickBooks Online"


class SamlLoginProvider(Enum):
    GOOGLE = "Google"
    RIPPLING = "Rippling"
    ONELOGIN = "OneLogin"
    JUMPCLOUD = "JumpCloud"
    OKTA = "Okta"
    AZURE = "Azure"
    DUO = "Duo"
    GENERIC = "Generic"


@dataclass
class PasswordLoginMethod:
    login_method: str = "password"
    provider = None
    org_id = None
    
    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class MagicLinkLoginMethod:
    login_method: str = "magic_link"
    provider = None
    org_id = None
    
    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class SocialSsoLoginMethod:
    provider: SocialLoginProvider
    login_method: str = "social_sso"
    org_id = None

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class EmailConfirmationLinkLoginMethod:
    login_method: str = "email_confirmation_link"
    provider = None
    org_id = None
    
    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class SamlSsoLoginMethod:
    provider: SamlLoginProvider
    org_id: str
    login_method: str = "saml_sso"

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class ImpersonationLoginMethod:
    login_method: str = "impersonation"
    provider = None
    org_id = None
    
    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class GeneratedFromBackendApiLoginMethod:
    login_method: str = "generated_from_backend_api"
    provider = None
    org_id = None
    
    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class UnknownLoginMethod:
    login_method: str = "unknown"
    provider = None
    org_id = None
    
    def __getitem__(self, key):
        return getattr(self, key)


def to_login_method(d={}):
    if "login_method" not in d:
        return UnknownLoginMethod()
    if d["login_method"] == "password":
        return PasswordLoginMethod()
    if d["login_method"] == "magic_link":
        return MagicLinkLoginMethod()
    if d["login_method"] == "social_sso":
        return SocialSsoLoginMethod(
            provider=SocialLoginProvider(d["provider"]),
        )
    if d["login_method"] == "email_confirmation_link":
        return EmailConfirmationLinkLoginMethod()
    if d["login_method"] == "saml_sso":
        return SamlSsoLoginMethod(
            provider=SamlLoginProvider(d["provider"]),
            org_id=d["org_id"],
        )
    if d["login_method"] == "impersonation":
        return ImpersonationLoginMethod()
    if d["login_method"] == "generated_from_backend_api":
        return GeneratedFromBackendApiLoginMethod()
    return UnknownLoginMethod()



class CreateUserException(Exception):
    def __init__(self, field_to_errors):
        self.field_to_errors = field_to_errors


class UpdateUserMetadataException(Exception):
    def __init__(self, field_to_errors):
        self.field_to_errors = field_to_errors


class InviteUserToOrgException(Exception):
    def __init__(self, field_to_errors):
        self.field_to_errors = field_to_errors


class UpdateUserPasswordException(Exception):
    def __init__(self, field_to_errors):
        self.field_to_errors = field_to_errors


class UpdateUserEmailException(Exception):
    def __init__(self, field_to_errors):
        self.field_to_errors = field_to_errors


class BadRequestException(Exception):
    def __init__(self, field_to_errors):
        self.field_to_errors = field_to_errors


class UserNotFoundException(Exception):
    pass


class EndUserApiKeyException(Exception):
    def __init__(self, field_to_errors):
        self.field_to_errors = field_to_errors


class EndUserApiKeyNotFoundException(Exception):
    pass


class UnauthorizedException(Exception):
    def __init__(self, message):
        self.message = message

    @staticmethod
    def no_header_found():
        return UnauthorizedException("No authorization header found")

    @staticmethod
    def invalid_header_found():
        return UnauthorizedException(
            "Invalid authorization header. Expected: Bearer {accessToken}"
        )

    @staticmethod
    def invalid_access_token():
        return UnauthorizedException("Invalid access token")

    @staticmethod
    def invalid_payload_in_access_token():
        return UnauthorizedException("Invalid payload in token")


class ForbiddenException(Exception):
    def __init__(self, message):
        self.message = message

    @staticmethod
    def unknown_required_org():
        return ForbiddenException("Required org is unspecified")

    @staticmethod
    def user_not_member_of_org(org_id):
        return ForbiddenException("User is not a member of org {}".format(org_id))

    @staticmethod
    def user_doesnt_have_required_role():
        return ForbiddenException("User doesn't have required role")

    @staticmethod
    def user_doesnt_have_required_permission():
        return ForbiddenException("User doesn't have required permission")



import requests
from propelauth_py.api import _ApiKeyAuth, _is_valid_id
from propelauth_py.errors import BadRequestException, UserNotFoundException

ENDPOINT_PATH = "/api/backend/v1/access_token"

class CreateAccessTokenResponse:
    def __init__(
        self,
        access_token: str,
    ):
        self.access_token = access_token

    def __repr__(self): 
        return (
            f"CreateAccessTokenResponse(access_token={self.access_token}"
        )
    def __eq__(self, other):
        return isinstance(other, CreateAccessTokenResponse)

def _create_access_token(auth_url, integration_api_key, user_id, duration_in_minutes) -> CreateAccessTokenResponse:
    if not _is_valid_id(user_id):
        raise UserNotFoundException()

    url = auth_url + ENDPOINT_PATH
    json = {"user_id": user_id, "duration_in_minutes": duration_in_minutes}
    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 403:
        raise UserNotFoundException()
    elif response.status_code == 404:
        raise RuntimeError("Access token creation is not enabled")
    elif not response.ok:
        raise RuntimeError("Unknown error when creating access token")

    json_response = response.json()
    return CreateAccessTokenResponse(
        access_token=json_response.get('access_token')
    )



from typing import Optional
import requests
from propelauth_py.api import _ApiKeyAuth, _is_valid_hex, remove_bearer_if_exists
from propelauth_py.errors import EndUserApiKeyException, EndUserApiKeyNotFoundException
from propelauth_py.types.end_user_api_keys import ApiKeyFull, ApiKeyResultPage, ApiKeyNew, ApiKeyValidation
from propelauth_py.types.user import UserMetadata, Org
from propelauth_py.user import OrgMemberInfo

ENDPOINT_PATH = "/api/backend/v1/end_user_api_keys"


####################
#       GET        #
####################
def _fetch_api_key(auth_url, integration_api_key, api_key_id) -> ApiKeyFull:
    if not _is_valid_hex(api_key_id):
        raise EndUserApiKeyNotFoundException()

    url = auth_url + f"{ENDPOINT_PATH}/{api_key_id}"    
    response = requests.get(url, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif response.status_code == 404:
        raise EndUserApiKeyNotFoundException()
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching end user api key")

    json_response = response.json()
    return ApiKeyFull(
        api_key_id=json_response.get('api_key_id'),
        created_at=json_response.get('created_at'),
        expires_at_seconds=json_response.get('expires_at_seconds'),
        metadata=json_response.get('metadata'),
        user_id=json_response.get('user_id'),
        org_id=json_response.get('org_id')
    )


def _fetch_current_api_keys(
    auth_url,
    integration_api_key,
    org_id=None,
    user_id=None,
    user_email=None,
    page_size=None,
    page_number=None,
    api_key_type=None,
) -> ApiKeyResultPage:
    url = auth_url + ENDPOINT_PATH

    query_params = {}
    if org_id:
        query_params["org_id"] = org_id
    if user_id:
        query_params["user_id"] = user_id
    if user_email:
        query_params["user_email"] = user_email
    if page_size:
        query_params["page_size"] = page_size
    if page_number:
        query_params["page_number"] = page_number
    if api_key_type:
        query_params["api_key_type"] = api_key_type

    response = requests.get(
        url, auth=_ApiKeyAuth(integration_api_key), params=query_params
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching current end user api keys")

    json_response = response.json()
    
    api_keys = [
        ApiKeyFull(
            api_key_id=key.get("api_key_id"),
            created_at=key.get("created_at"),
            expires_at_seconds=key.get("expires_at_seconds"),
            metadata=key.get("metadata"),
            user_id=key.get("user_id"),
            org_id=key.get("org_id")
        )
        for key in json_response.get('api_keys')
    ]
    
    return ApiKeyResultPage(
        api_keys=api_keys,
        total_api_keys=json_response.get('total_api_keys'),
        current_page=json_response.get('current_page'),
        page_size=json_response.get('page_size'),
        has_more_results=json_response.get('has_more_results')
    )


def _fetch_archived_api_keys(
    auth_url,
    integration_api_key,
    org_id=None,
    user_id=None,
    user_email=None,
    page_size=None,
    page_number=None,
    api_key_type=None,
) -> ApiKeyResultPage:
    url = auth_url + f"{ENDPOINT_PATH}/archived"

    query_params = {}
    if org_id:
        query_params["org_id"] = org_id
    if user_id:
        query_params["user_id"] = user_id
    if user_email:
        query_params["user_email"] = user_email
    if page_size:
        query_params["page_size"] = page_size
    if page_number:
        query_params["page_number"] = page_number
    if api_key_type:
        query_params["api_key_type"] = api_key_type

    response = requests.get(
        url, auth=_ApiKeyAuth(integration_api_key), params=query_params
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching archived end user api keys")

    json_response = response.json()
    
    api_keys = [
        ApiKeyFull(
            api_key_id=key.get("api_key_id"),
            created_at=key.get("created_at"),
            expires_at_seconds=key.get("expires_at_seconds"),
            metadata=key.get("metadata"),
            user_id=key.get("user_id"),
            org_id=key.get("org_id")
        )
        for key in json_response.get('api_keys')
    ]
    
    return ApiKeyResultPage(
        api_keys=api_keys,
        total_api_keys=json_response.get('total_api_keys'),
        current_page=json_response.get('current_page'),
        page_size=json_response.get('page_size'),
        has_more_results=json_response.get('has_more_results')
    )


####################
#       POST       #
####################
def _create_api_key(
    auth_url, integration_api_key, org_id, user_id, expires_at_seconds, metadata
) -> ApiKeyNew:
    url = auth_url + ENDPOINT_PATH

    json = {}
    if org_id:
        json["org_id"] = org_id
    if user_id:
        json["user_id"] = user_id
    if expires_at_seconds:
        json["expires_at_seconds"] = expires_at_seconds
    if metadata:
        json["metadata"] = metadata

    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key), json=json)

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when creating end user api key")

    json_response = response.json()
    return ApiKeyNew(
        api_key_id=json_response.get('api_key_id'),
        api_key_token=json_response.get('api_key_token')
    )


def _validate_api_key(auth_url, integration_api_key, api_key_token) -> ApiKeyValidation:
    url = auth_url + f"{ENDPOINT_PATH}/validate"
    json = {"api_key_token": remove_bearer_if_exists(api_key_token)}
    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key), json=json)

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif response.status_code == 404:
        raise EndUserApiKeyNotFoundException()
    elif not response.ok:
        raise RuntimeError("Unknown error when validating end user api key")

    json_response = response.json()
    
    user = None
    if json_response.get('user') is not None:
        user_data = json_response.get('user')
        user = UserMetadata(
            user_id=user_data.get('user_id'),
            email=user_data.get('email'),
            email_confirmed=user_data.get('email_confirmed'),
            has_password=user_data.get('has_password'),
            username=user_data.get('username'),
            first_name=user_data.get('first_name'),
            last_name=user_data.get('last_name'),
            picture_url=user_data.get('picture_url'),
            locked=user_data.get('locked'),
            enabled=user_data.get('enabled'),
            mfa_enabled=user_data.get('mfa_enabled'),
            can_create_orgs=user_data.get('can_create_orgs'),
            created_at=user_data.get('created_at'),
            last_active_at=user_data.get('last_active_at'),
            org_id_to_org_info=user_data.get('org_id_to_org_info'),
            legacy_org_id=user_data.get('legacy_org_id'),
            impersonator_user_id=user_data.get('impersonator_user_id'),
            metadata=user_data.get('metadata'),
            properties=user_data.get('properties')
        )
        
    org = None
    if json_response.get('org') is not None:
        org_data = json_response.get('org')
        org = Org(
            org_id=org_data.get('org_id'),
            name=org_data.get('org_name'),
            max_users=org_data.get('max_users'),
            is_saml_configured=org_data.get('is_saml_configured'),
            legacy_org_id=org_data.get('legacy_org_id'),
            metadata=org_data.get('metadata', {}),
            custom_role_mapping_name=org_data.get('custom_role_mapping_name')
        )
        
    user_in_org = None
    if json_response.get('user_in_org') is not None:
        user_in_org_data = json_response.get('user_in_org')
        user_in_org = OrgMemberInfo(
            org_id=user_in_org_data.get('org_id'),
            org_name=user_in_org_data.get('org_name'),
            org_metadata=user_in_org_data.get('org_metadata'),
            user_assigned_role=user_in_org_data.get('user_role'),
            url_safe_org_name=user_in_org_data.get('url_safe_org_name'),
            user_inherited_roles_plus_current_role=user_in_org_data.get('inherited_user_roles_plus_current_role'),
            user_permissions=user_in_org_data.get('user_permissions'),
            org_role_structure=user_in_org_data.get('org_role_structure'),
            assigned_additional_roles=user_in_org_data.get('additional_roles', [])
        )
    
    return ApiKeyValidation(
        metadata=json_response.get('metadata'),
        user=user,
        org=org,
        user_in_org=user_in_org,
    )


####################
#    PUT/PATCH     #
####################
def _update_api_key(
    auth_url, integration_api_key, api_key_id, expires_at_seconds, metadata
) -> bool:
    if not _is_valid_hex(api_key_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{api_key_id}"

    json = {}
    if expires_at_seconds:
        json["expires_at_seconds"] = expires_at_seconds
    if metadata:
        json["metadata"] = metadata

    response = requests.patch(url, auth=_ApiKeyAuth(integration_api_key), json=json)

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif response.status_code == 404:
        raise EndUserApiKeyNotFoundException()
    elif not response.ok:
        raise RuntimeError("Unknown error when updating end user api key")

    return True


####################
#      DELETE      #
####################
def _delete_api_key(auth_url, integration_api_key, api_key_id) -> bool:
    if not _is_valid_hex(api_key_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{api_key_id}"
    response = requests.delete(url, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise EndUserApiKeyException(response.json())
    elif response.status_code == 404:
        raise EndUserApiKeyNotFoundException()
    elif not response.ok:
        raise RuntimeError("Unknown error when deleting end user api key")

    return True



from typing import Optional, Any, Dict, Union
from propelauth_py.errors import UnauthorizedException
from propelauth_py.types.login_method import (
    to_login_method,
    PasswordLoginMethod,
    MagicLinkLoginMethod,
    SocialSsoLoginMethod,
    EmailConfirmationLinkLoginMethod,
    SamlSsoLoginMethod,
    ImpersonationLoginMethod,
    GeneratedFromBackendApiLoginMethod,
    UnknownLoginMethod
)
from dataclasses import dataclass, field

MULTI_ROLE = "multi_role"
SINGLE_ROLE = "single_role_in_hierarchy"

@dataclass
class OrgMemberInfo:
    org_id: str
    org_name: str
    org_metadata: Dict[str, Any]
    user_assigned_role: str
    url_safe_org_name: str
    user_inherited_roles_plus_current_role: list[str]
    user_permissions: list[str]
    org_role_structure: str = SINGLE_ROLE
    assigned_additional_roles: list[str] = field(default_factory=list)
    
    def __getitem__(self, key):
        return getattr(self, key)

    def user_is_role(self, role: str) -> bool:
        """returns true if the user is the role"""
        return (role == self.user_assigned_role or 
            (self.org_role_structure == MULTI_ROLE and role in self.assigned_additional_roles))

    def user_is_at_least_role(self, role: str) -> bool:
        """returns true if the user can act as the role"""
        if self.org_role_structure == MULTI_ROLE:
            return role == self.user_assigned_role or role in self.assigned_additional_roles
        else:
            return role in self.user_inherited_roles_plus_current_role

    def user_has_permission(self, permission: str) -> bool:
        """returns true if user has the permission"""
        return permission in self.user_permissions

    def user_has_all_permissions(self, permissions: list[str]) -> bool:
        """returns true if user has all the permissions listed"""
        for permission in permissions:
            if not self.user_has_permission(permission):
                return False

        return True

@dataclass
class User:
    user_id: str
    org_id_to_org_member_info: Optional[Dict[str, OrgMemberInfo]]
    email: str
    login_method: Union[
        PasswordLoginMethod, 
        MagicLinkLoginMethod, 
        SocialSsoLoginMethod, 
        EmailConfirmationLinkLoginMethod, 
        SamlSsoLoginMethod, 
        ImpersonationLoginMethod, 
        GeneratedFromBackendApiLoginMethod, 
        UnknownLoginMethod
    ]
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    username: Optional[str] = None
    properties: Optional[Dict[str, Any]] = None
    legacy_user_id: Optional[str] = None
    impersonator_user_id: Optional[str] = None
    active_org_id: Optional[str] = None

    def __getitem__(self, key):
        return getattr(self, key)

    def is_impersonated(self) -> bool:
        """Returns true if the user is impersonated"""
        return self.impersonator_user_id is not None

    def get_active_org(self) -> Optional[OrgMemberInfo]:
        """Returns the active org member info, if the user has an active org."""
        if self.active_org_id is None:
            return None
        return self.get_org(self.active_org_id)

    def get_active_org_id(self) -> Optional[str]:
        """Returns the active org id, if the user has an active org."""
        return self.active_org_id

    def get_org(self, org_id: str) -> Optional[OrgMemberInfo]:
        """Returns the org member info for the org_id, if the user is in the org."""
        if self.org_id_to_org_member_info:
            return self.org_id_to_org_member_info.get(org_id)
        return None

    def get_org_by_name(self, org_name: str) -> Optional[OrgMemberInfo]:
        """Returns the org member info for the org_name, if the user is in the org."""
        if self.org_id_to_org_member_info:
            for org_member_info in self.org_id_to_org_member_info.values():
                if org_member_info.org_name == org_name:
                    return org_member_info
        return None

    def get_user_property(self, property_name: str):
        """Returns the user property value, if it exists."""
        if self.properties:
            return self.properties.get(property_name)
        return None

    def get_orgs(self) -> list[OrgMemberInfo]:
        """Returns the orgs the user is in."""
        if self.org_id_to_org_member_info:
            return list(self.org_id_to_org_member_info.values())
        return []

    def is_role_in_org(self, org_id: str, role: str) -> bool:
        """Returns true if the user is the role in the org."""
        org_member_info = self.get_org(org_id)
        if not org_member_info:
            return False

        return org_member_info.user_is_role(role)

    def is_at_least_role_in_org(self, org_id: str, role: str) -> bool:
        """Returns true if the user is at least the role in the org."""
        org_member_info = self.get_org(org_id)
        if not org_member_info:
            return False

        return org_member_info.user_is_at_least_role(role)

    def has_permission_in_org(self, org_id: str, permission: str) -> bool:
        """Returns true if the user has the permission in the org."""
        org_member_info = self.get_org(org_id)
        if not org_member_info:
            return False

        return org_member_info.user_has_permission(permission)

    def has_all_permissions_in_org(self, org_id: str, permissions: list[str]) -> bool:
        """Returns true if the user has all the permissions in the org."""
        org_member_info = self.get_org(org_id)
        if not org_member_info:
            return False

        return org_member_info.user_has_all_permissions(permissions)


@dataclass
class UserAndOrgMemberInfo:
    user: User
    org_member_info: OrgMemberInfo

    def __getitem__(self, key):
        return getattr(self, key)


def _to_org_member_info(org_id_to_org_member_info_json) -> Optional[Dict[str, OrgMemberInfo]]:
    if org_id_to_org_member_info_json is None:
        return None

    org_id_to_org_member_info = {}
    for org_id, org_member_info_json in org_id_to_org_member_info_json.items():
        user_assigned_role = org_member_info_json["user_role"]
        if user_assigned_role is not None:
            org_id_to_org_member_info[org_id] = OrgMemberInfo(
                org_id=org_member_info_json["org_id"],
                org_name=org_member_info_json["org_name"],
                org_metadata=org_member_info_json["org_metadata"],
                user_assigned_role=user_assigned_role,
                url_safe_org_name=org_member_info_json["url_safe_org_name"],
                user_inherited_roles_plus_current_role=org_member_info_json[
                    "inherited_user_roles_plus_current_role"
                ],
                user_permissions=org_member_info_json["user_permissions"],
                org_role_structure=org_member_info_json.get("org_role_structure", SINGLE_ROLE),
                assigned_additional_roles=org_member_info_json.get("additional_roles", []),
            )
    return org_id_to_org_member_info


def _to_user(decoded_token):
    user_id = decoded_token.get("user_id")
    if user_id is None:
        raise UnauthorizedException.invalid_payload_in_access_token()

    org_member_info = decoded_token.get("org_member_info")
    if org_member_info:
        active_org_id = org_member_info.get("org_id")
        org_id_to_org_member_info = _to_org_member_info(
            {active_org_id: org_member_info}
        )
    else:
        active_org_id = None
        org_id_to_org_member_info = _to_org_member_info(
            decoded_token.get("org_id_to_org_member_info")
        )

    return User(
        user_id,
        org_id_to_org_member_info,
        decoded_token.get("email"),
        first_name=decoded_token.get("first_name"),
        last_name=decoded_token.get("last_name"),
        username=decoded_token.get("username"),
        legacy_user_id=decoded_token.get("legacy_user_id"),
        impersonator_user_id=decoded_token.get("impersonator_user_id"),
        properties=decoded_token.get("properties"),
        active_org_id=active_org_id,
        login_method=to_login_method(decoded_token.get("login_method", {})),
    )


from dataclasses import dataclass

@dataclass
class SpMetadata:
    entity_id: str
    acs_url: str
    logout_url: str

class SamlIdpMetadata(dict):
    def __init__(self, idp_entity_id: str, idp_sso_url: str, idp_certificate: str, provider: str):
        super().__init__()
        self["idp_entity_id"] = idp_entity_id
        self["idp_sso_url"] = idp_sso_url
        self["idp_certificate"] = idp_certificate
        self["provider"] = provider

    @property
    def idp_entity_id(self):
        return self["idp_entity_id"]

    @property
    def idp_sso_url(self):
        return self["idp_sso_url"]

    @property
    def idp_certificate(self):
        return self["idp_certificate"]

    @property
    def provider(self):
        return self["provider"]





from typing import Any, Dict, Optional
from propelauth_py.user import OrgMemberInfo
from dataclasses import dataclass

OrgIdToOrgMemberInfo = Dict[str, OrgMemberInfo]

@dataclass
class Org:
    org_id: str
    name: str
    max_users: Optional[int]
    is_saml_configured: bool
    legacy_org_id: Optional[str]
    metadata: Dict[str, Any]
    custom_role_mapping_name: Optional[str]

    def __getitem__(self, key):
        return getattr(self, key)
    
@dataclass
class Organization:
    org_id: str
    name: str
    url_safe_org_slug: str
    can_setup_saml: bool
    is_saml_configured: bool
    is_saml_in_test_mode: bool
    max_users: Optional[int]
    metadata: Optional[Dict[str, Any]]
    domain: Optional[str]
    domain_autojoin: bool
    domain_restrict: bool
    custom_role_mapping_name: Optional[str]
    legacy_org_id: Optional[str]

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class OrgQueryResponse:
    orgs: list[Org]
    total_orgs: int
    current_page: int
    page_size: int
    has_more_results: bool

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class PendingInvite:
    invitee_email: str
    org_id: str
    org_name: str
    role_in_org: str
    additional_roles_in_org: list[str]
    created_at: int
    expires_at: int
    inviter_email: Optional[str]
    inviter_user_id: Optional[str]

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class PendingInvitesPage:
    total_invites: int
    current_page: int
    page_size: int
    has_more_results: bool
    invites: list[PendingInvite]

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class CreatedOrg:
    org_id: str
    name: str

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class UserMetadata:
    user_id: str
    email: str
    email_confirmed: bool
    has_password: bool
    username: Optional[str]
    first_name: Optional[str]
    last_name: Optional[str]
    picture_url: Optional[str]
    locked: bool
    enabled: bool
    mfa_enabled: bool
    can_create_orgs: bool
    created_at: int
    last_active_at: int
    org_id_to_org_info: Optional[OrgIdToOrgMemberInfo]
    legacy_org_id: Optional[str]
    impersonator_user_id: Optional[str]
    metadata: Optional[Dict[str, Any]]
    properties: Optional[Dict[str, Any]]

    def __getitem__(self, key):
        return getattr(self, key)
    
@dataclass
class OrgApiKeyValidation:
    metadata: Optional[Dict[str, Any]]
    org: Org
    user: Optional[UserMetadata]
    user_in_org: Optional[OrgMemberInfo]

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass    
class CreatedUser:
    user_id: str

    def __getitem__(self, key):
        return getattr(self, key)
    
    
@dataclass
class UsersPagedResponse:
    users: list[UserMetadata]
    total_users: int
    current_page: int
    page_size: int
    has_more_results: bool

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class PersonalApiKeyValidation:
    metadata: Optional[Dict[str, Any]]
    user: UserMetadata

    def __getitem__(self, key):
        return getattr(self, key)

@dataclass
class UserSignupQueryParams:
    user_signup_query_parameters: Dict[str, str]

    def __getitem__(self, key):
        return getattr(self, key)



import datetime

import jwt

from propelauth_py.errors import UnauthorizedException
from propelauth_py.user import _to_user

OPTIONS = {
    "verify_signature": True,
    "verify_exp": True,
    "verify_iat": True,
    "verify_iss": True,
    "verify_aud": False,
    "require": ["exp", "iat", "iss"],
}


def _validate_access_token_and_get_user(access_token, token_verification_metadata):
    try:
        decoded_token = jwt.decode(access_token,
                                   token_verification_metadata.verifier_key,
                                   options=OPTIONS,
                                   issuer=token_verification_metadata.issuer,
                                   algorithms=["RS256"],
                                   leeway=datetime.timedelta(seconds=60))
        return _to_user(decoded_token)
    except UnauthorizedException as e:
        raise e
    except Exception:
        raise UnauthorizedException.invalid_access_token()



from typing import Optional, Dict
import requests

from propelauth_py.api import _ApiKeyAuth, _format_params, _is_valid_id
from propelauth_py.api.end_user_api_keys import _validate_api_key
from propelauth_py.types.user import UserMetadata, UsersPagedResponse, CreatedUser, PersonalApiKeyValidation, UserSignupQueryParams
from propelauth_py.errors import (
    CreateUserException,
    EndUserApiKeyException,
    InviteUserToOrgException,
    UpdateUserEmailException,
    UpdateUserMetadataException,
    UpdateUserPasswordException,
)

ENDPOINT_PATH = "/api/backend/v1/user"


####################
#       GET        #
####################
def _fetch_user_metadata_by_user_id(
    auth_url, integration_api_key, user_id, include_orgs=False
) -> Optional[UserMetadata]:
    if not _is_valid_id(user_id):
        return None

    user_info_url = auth_url + f"{ENDPOINT_PATH}/{user_id}"
    query = {"include_orgs": include_orgs}
    return _fetch_user_metadata_by_query(integration_api_key, user_info_url, query)


def _fetch_user_signup_query_params_by_user_id(
    auth_url,
    integration_api_key,
    user_id,
) -> Optional[UserSignupQueryParams]:
    if not _is_valid_id(user_id):
        return None

    user_signup_query_params_url = (
        auth_url + f"{ENDPOINT_PATH}/{user_id}/signup_query_parameters"
    )
    response = requests.get(
        url=user_signup_query_params_url, auth=_ApiKeyAuth(integration_api_key)
    )

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return None
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching user signup query params")

    json_response = response.json()
    return UserSignupQueryParams(
        user_signup_query_parameters=json_response.get('user_signup_query_parameters')
    )


def _fetch_user_metadata_by_email(
    auth_url, integration_api_key, email, include_orgs=False
) -> Optional[UserMetadata]:
    user_info_url = auth_url + f"{ENDPOINT_PATH}/email"
    query = {"include_orgs": include_orgs, "email": email}
    return _fetch_user_metadata_by_query(integration_api_key, user_info_url, query)


def _fetch_user_metadata_by_username(
    auth_url, integration_api_key, username, include_orgs=False
) -> Optional[UserMetadata]:
    user_info_url = auth_url + f"{ENDPOINT_PATH}/username"
    query = {"include_orgs": include_orgs, "username": username}
    return _fetch_user_metadata_by_query(integration_api_key, user_info_url, query)


def _fetch_user_metadata_by_query(integration_api_key, user_info_url, query) -> Optional[UserMetadata]:
    response = requests.get(
        user_info_url,
        params=_format_params(query),
        auth=_ApiKeyAuth(integration_api_key),
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif response.status_code == 404:
        return None
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching user metadata")

    json_response = response.json()

    return UserMetadata(
        user_id=json_response.get('user_id'),
        email=json_response.get('email'),
        email_confirmed=json_response.get('email_confirmed'),
        has_password=json_response.get('has_password'),
        username=json_response.get('username'),
        first_name=json_response.get('first_name'),
        last_name=json_response.get('last_name'),
        picture_url=json_response.get('picture_url'),
        locked=json_response.get('locked'),
        enabled=json_response.get('enabled'),
        mfa_enabled=json_response.get('mfa_enabled'),
        can_create_orgs=json_response.get('can_create_orgs'),
        created_at=json_response.get('created_at'),
        last_active_at=json_response.get('last_active_at'),
        org_id_to_org_info=json_response.get('org_id_to_org_info'),
        legacy_org_id=json_response.get('legacy_org_id'),
        impersonator_user_id=json_response.get('impersonator_user_id'),
        metadata=json_response.get('metadata'),
        properties=json_response.get('properties')
    )


def _fetch_batch_user_metadata_by_user_ids(
    auth_url, integration_api_key, user_ids, include_orgs
) -> Dict[str, UserMetadata]:
    user_info_url = auth_url + f"{ENDPOINT_PATH}/user_ids"
    params = {"include_orgs": include_orgs}
    body = {"user_ids": user_ids}
    return _fetch_batch_user_metadata_by_query(
        user_info_url, integration_api_key, params, body, lambda x: x["user_id"]
    )


def _fetch_batch_user_metadata_by_emails(
    auth_url, integration_api_key, emails, include_orgs
) -> Dict[str, UserMetadata]:
    user_info_url = auth_url + f"{ENDPOINT_PATH}/emails"
    params = {"include_orgs": include_orgs}
    body = {"emails": emails}
    return _fetch_batch_user_metadata_by_query(
        user_info_url, integration_api_key, params, body, lambda x: x["email"]
    )


def _fetch_batch_user_metadata_by_usernames(
    auth_url, integration_api_key, usernames, include_orgs
) -> Dict[str, UserMetadata]:
    user_info_url = auth_url + f"{ENDPOINT_PATH}/usernames"
    params = {"include_orgs": include_orgs}
    body = {"usernames": usernames}
    return _fetch_batch_user_metadata_by_query(
        user_info_url, integration_api_key, params, body, lambda x: x["username"]
    )


def _fetch_batch_user_metadata_by_query(
    user_info_url, integration_api_key, params, body, key_fn
) -> Dict[str, UserMetadata]:
    response = requests.post(
        user_info_url,
        params=_format_params(params),
        json=body,
        auth=_ApiKeyAuth(integration_api_key),
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching batch user metadata")

    json_response = response.json()
    return_value = {}
    for single_item in json_response:
        return_value[key_fn(single_item)] = single_item

    return return_value


def _fetch_users_by_query(
    auth_url,
    integration_api_key,
    page_size,
    page_number,
    order_by,
    email_or_username,
    include_orgs,
    legacy_user_id,
) -> UsersPagedResponse:
    url = auth_url + f"{ENDPOINT_PATH}/query"
    params = {
        "page_size": page_size,
        "page_number": page_number,
        "order_by": order_by,
        "email_or_username": email_or_username,
        "include_orgs": include_orgs,
        "legacy_user_id": legacy_user_id,
    }
    response = requests.get(
        url, params=_format_params(params), auth=_ApiKeyAuth(integration_api_key)
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching orgs by query")

    json_response = response.json()
    
    users = [
        UserMetadata(
            user_id=key.get('user_id'),
            email=key.get('email'),
            email_confirmed=key.get('email_confirmed'),
            has_password=key.get('has_password'),
            username=key.get('username'),
            first_name=key.get('first_name'),
            last_name=key.get('last_name'),
            picture_url=key.get('picture_url'),
            locked=key.get('locked'),
            enabled=key.get('enabled'),
            mfa_enabled=key.get('mfa_enabled'),
            can_create_orgs=key.get('can_create_orgs'),
            created_at=key.get('created_at'),
            last_active_at=key.get('last_active_at'),
            org_id_to_org_info=key.get('org_id_to_org_info'),
            legacy_org_id=key.get('legacy_org_id'),
            impersonator_user_id=key.get('impersonator_user_id'),
            metadata=key.get('metadata'),
            properties=key.get('properties')
        )
        for key in json_response.get('users')
    ]
    
    return UsersPagedResponse(
        users=users,
        total_users=json_response.get('total_users'),
        current_page=json_response.get('current_page'),
        page_size=json_response.get('page_size'),
        has_more_results=json_response.get('has_more_results')
    )


def _fetch_users_in_org(
    auth_url, integration_api_key, org_id, page_size, page_number, include_orgs, role
) -> UsersPagedResponse:
    if not _is_valid_id(org_id):
        return UsersPagedResponse(
            users=[],
            total_users=0,
            current_page=page_number,
            page_size=page_size,
            has_more_results=False,
        )

    url = auth_url + f"{ENDPOINT_PATH}/org/{org_id}"
    params = {
        "page_size": page_size,
        "page_number": page_number,
        "include_orgs": include_orgs,
        "role": role,
    }
    response = requests.get(
        url, params=_format_params(params), auth=_ApiKeyAuth(integration_api_key)
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching users in org")

    json_response = response.json()
    
    users = [
        UserMetadata(
            user_id=key.get('user_id'),
            email=key.get('email'),
            email_confirmed=key.get('email_confirmed'),
            has_password=key.get('has_password'),
            username=key.get('username'),
            first_name=key.get('first_name'),
            last_name=key.get('last_name'),
            picture_url=key.get('picture_url'),
            locked=key.get('locked'),
            enabled=key.get('enabled'),
            mfa_enabled=key.get('mfa_enabled'),
            can_create_orgs=key.get('can_create_orgs'),
            created_at=key.get('created_at'),
            last_active_at=key.get('last_active_at'),
            org_id_to_org_info=key.get('org_id_to_org_info'),
            legacy_org_id=key.get('legacy_org_id'),
            impersonator_user_id=key.get('impersonator_user_id'),
            metadata=key.get('metadata'),
            properties=key.get('properties')
        )
        for key in json_response.get('users')
    ]
    
    return UsersPagedResponse(
        users=users,
        total_users=json_response.get('total_users'),
        current_page=json_response.get('current_page'),
        page_size=json_response.get('page_size'),
        has_more_results=json_response.get('has_more_results')
    )


####################
#       POST     #
####################
def _create_user(
    auth_url,
    integration_api_key,
    email,
    email_confirmed,
    send_email_to_confirm_email_address,
    ask_user_to_update_password_on_login,
    password,
    username,
    first_name,
    last_name,
    properties,
) -> CreatedUser:
    url = auth_url + f"{ENDPOINT_PATH}/"
    json = {
        "email": email,
        "email_confirmed": email_confirmed,
        "send_email_to_confirm_email_address": send_email_to_confirm_email_address,
        "ask_user_to_update_password_on_login": ask_user_to_update_password_on_login,
    }
    if password is not None:
        json["password"] = password
    if username is not None:
        json["username"] = username
    if first_name is not None:
        json["first_name"] = first_name
    if last_name is not None:
        json["last_name"] = last_name
    if properties is not None:
        json["properties"] = properties
    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise CreateUserException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when creating user")

    json_response = response.json()
    return CreatedUser(
        user_id=json_response.get('user_id')
    )


def _disable_user(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/disable"
    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when disabling user")

    return True


def _enable_user(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/enable"
    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when enabling user")

    return True


def _disable_user_2fa(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/disable_2fa"
    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when enabling user")

    return True


def _invite_user_to_org(
    auth_url, integration_api_key, email, org_id, role, additional_roles=[]
) -> bool:
    if not _is_valid_id(org_id):
        return False

    endpoint_path = "/api/backend/v1/invite_user"
    url = auth_url + endpoint_path
    json = {
        "email": email,
        "org_id": org_id,
        "role": role,
        "additional_roles": additional_roles,
    }
    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        try:
            response_json = response.json()
            raise InviteUserToOrgException(response_json)
        except requests.exceptions.JSONDecodeError:
            raise ValueError("Bad request: " + response.text)
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating metadata")

    return True


def _resend_email_confirmation(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    endpoint_path = "/api/backend/v1/resend_email_confirmation"
    url = auth_url + endpoint_path
    json = {
        "user_id": user_id,
    }
    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif response.status_code == 429:
        raise RuntimeError("Too many requests, please try again later.")
    elif response.status_code == 400:
        if response.json().get("user_facing_error"):
            raise ValueError(response.json().get("user_facing_error"))
        else:
            raise RuntimeError("Unknown error when resending email confirmation")
    elif not response.ok:
        raise RuntimeError("Unknown error when resending email confirmation")

    return True


def _logout_all_user_sessions(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/logout_all_sessions"
    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when logging out all user sessions")

    return True


####################
#     PATCH/PUT    #
####################
def _update_user_metadata(
    auth_url,
    integration_api_key,
    user_id,
    username=None,
    first_name=None,
    last_name=None,
    metadata=None,
    properties=None,
    picture_url=None,
    update_password_required=None,
)-> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}"
    json = {}
    if username is not None:
        json["username"] = username
    if first_name is not None:
        json["first_name"] = first_name
    if last_name is not None:
        json["last_name"] = last_name
    if metadata is not None:
        json["metadata"] = metadata
    if properties is not None:
        json["properties"] = properties
    if picture_url is not None:
        json["picture_url"] = picture_url
    if update_password_required is not None:
        json["update_password_required"] = update_password_required

    response = requests.put(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise UpdateUserMetadataException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating metadata")

    return True


def _update_user_password(
    auth_url,
    integration_api_key,
    user_id,
    password,
    ask_user_to_update_password_on_login,
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/password"
    json = {"password": password}
    if ask_user_to_update_password_on_login is not None:
        json["ask_user_to_update_password_on_login"] = (
            ask_user_to_update_password_on_login
        )

    response = requests.put(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise UpdateUserPasswordException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating password")

    return True


def _clear_user_password(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/clear_password"
    response = requests.put(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise UpdateUserEmailException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating user email")

    return True


def _update_user_email(
    auth_url, integration_api_key, user_id, new_email, require_email_confirmation
) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/email"
    json = {
        "new_email": new_email,
        "require_email_confirmation": require_email_confirmation,
    }

    response = requests.put(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise UpdateUserEmailException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating user email")

    return True


def _enable_user_can_create_orgs(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/can_create_orgs/enable"
    response = requests.put(url, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when enabling can_create_orgs")

    return True


def _disable_user_can_create_orgs(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}/can_create_orgs/disable"
    response = requests.put(url, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when disabling can_create_orgs")

    return True


####################
#       DELETE     #
####################
def _delete_user(auth_url, integration_api_key, user_id) -> bool:
    if not _is_valid_id(user_id):
        return False

    url = auth_url + f"{ENDPOINT_PATH}/{user_id}"
    response = requests.delete(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when deleting user")

    return True


####################
#       HELPERS    #
####################


def _validate_personal_api_key(auth_url, integration_api_key, api_key_token) -> PersonalApiKeyValidation:
    api_key_validation = _validate_api_key(auth_url, integration_api_key, api_key_token)
    if not api_key_validation.user or api_key_validation.org:
        raise EndUserApiKeyException({"api_key_token": ["Not a personal API Key"]})
    return PersonalApiKeyValidation(
        user=api_key_validation.user,
        metadata=api_key_validation.metadata,
    )



from typing import Optional
import requests
from propelauth_py.api import _ApiKeyAuth, TokenVerificationMetadata

ENDPOINT_PATH = "/api/v1/token_verification_metadata"


####################
#       GET        #
####################
def _fetch_token_verification_metadata(
    auth_url: str,
    integration_api_key: str,
    token_verification_metadata: Optional[TokenVerificationMetadata],
):
    if token_verification_metadata is not None:
        return token_verification_metadata

    token_verification_metadata_url = auth_url + ENDPOINT_PATH
    response = requests.get(
        token_verification_metadata_url, auth=_ApiKeyAuth(integration_api_key)
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise ValueError("Bad request")
    elif response.status_code == 404:
        raise ValueError("auth_url is incorrect")
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching token verification metadata")

    json_response = response.json()
    return TokenVerificationMetadata(
        verifier_key=json_response.get("verifier_key_pem"),
        issuer=auth_url,
    )



from typing import Optional, Any, Dict
from propelauth_py.user import UserAndOrgMemberInfo, User
from propelauth_py.jwt import _validate_access_token_and_get_user
from propelauth_py.api import token_verification_metadata, TokenVerificationMetadata, OrgQueryOrderBy, UserQueryOrderBy
from propelauth_py.api.user import (
    _clear_user_password,
    _fetch_user_metadata_by_user_id,
    _fetch_user_metadata_by_email,
    _fetch_user_metadata_by_username,
    _fetch_batch_user_metadata_by_user_ids,
    _fetch_batch_user_metadata_by_emails,
    _fetch_batch_user_metadata_by_usernames,
    _fetch_user_signup_query_params_by_user_id,
    _fetch_users_by_query,
    _fetch_users_in_org,
    _create_user,
    _logout_all_user_sessions,
    _update_user_email,
    _update_user_metadata,
    _delete_user,
    _disable_user,
    _enable_user,
    _update_user_password,
    _disable_user_2fa,
    _enable_user_can_create_orgs,
    _disable_user_can_create_orgs,
    _validate_personal_api_key,
    _invite_user_to_org,
    _resend_email_confirmation,

)
from propelauth_py.api.access_token import _create_access_token
from propelauth_py.api.end_user_api_keys import (
    _create_api_key,
    _delete_api_key,
    _fetch_api_key,
    _fetch_archived_api_keys,
    _fetch_current_api_keys,
    _update_api_key,
    _validate_api_key,
)
from propelauth_py.api.magic_link import _create_magic_link
from propelauth_py.api.migrate_user import _migrate_user_from_external_source
from propelauth_py.api.org import (
    _add_user_to_org,
    _allow_org_to_setup_saml_connection,
    _change_user_role_in_org,
    _create_org,
    _create_org_saml_connection_link,
    _delete_org,
    _disallow_org_to_setup_saml_connection,
    _fetch_custom_role_mappings,
    _fetch_org,
    _fetch_org_by_query,
    _fetch_pending_invites,
    _remove_user_from_org,
    _revoke_pending_org_invite,
    _subscribe_org_to_role_mapping,
    _update_org_metadata,
    _validate_org_api_key,
    _fetch_saml_sp_metadata,
    _set_saml_idp_metadata,
    _saml_go_live,
    _delete_saml_connection,
)
from propelauth_py.api.token_verification_metadata import (
    _fetch_token_verification_metadata,
)

from propelauth_py.auth_fns import (
    validate_all_permissions_and_get_org,
    validate_exact_org_role_and_get_org,
    validate_minimum_org_role_and_get_org,
    validate_org_access_and_get_org_member_info,
    validate_permission_and_get_org,
    _extract_token_from_authorization_header,
    wrap_validate_access_token_and_get_user,
    wrap_validate_access_token_and_get_user_with_org,
    wrap_validate_access_token_and_get_user_with_org_by_all_permissions,
    wrap_validate_access_token_and_get_user_with_org_by_exact_role,
    wrap_validate_access_token_and_get_user_with_org_by_minimum_role,
    wrap_validate_access_token_and_get_user_with_org_by_permission,
)
from propelauth_py.errors import ForbiddenException, UnauthorizedException
from propelauth_py.types.login_method import (
    EmailConfirmationLinkLoginMethod,
    GeneratedFromBackendApiLoginMethod,
    ImpersonationLoginMethod,
    MagicLinkLoginMethod,
    PasswordLoginMethod,
    SamlLoginProvider,
    SamlSsoLoginMethod,
    SocialLoginProvider,
    SocialSsoLoginMethod,
    UnknownLoginMethod,
)
from propelauth_py.types.saml_types import SamlIdpMetadata

from propelauth_py.validation import _validate_url

class Auth:
    
    def __init__(self, auth_url: str, integration_api_key: str, token_verification_metadata: Optional[TokenVerificationMetadata]):
        self.auth_url = auth_url
        self.integration_api_key = integration_api_key
        self.token_verification_metadata = token_verification_metadata

    def fetch_user_metadata_by_user_id(self, user_id: str, include_orgs: bool = False):
        return _fetch_user_metadata_by_user_id(self.auth_url, self.integration_api_key, user_id, include_orgs)

    def fetch_user_metadata_by_email(self, email: str, include_orgs: bool = False):
        return _fetch_user_metadata_by_email(self.auth_url, self.integration_api_key, email, include_orgs)
    
    def fetch_user_metadata_by_username(self, username: str, include_orgs: bool = False):
        return _fetch_user_metadata_by_username(
            self.auth_url, self.integration_api_key, username, include_orgs
        )

    def fetch_user_signup_query_params_by_user_id(self, user_id: str):
        return _fetch_user_signup_query_params_by_user_id(
            self.auth_url, self.integration_api_key, user_id
        )

    def fetch_batch_user_metadata_by_user_ids(self, user_ids: list[str], include_orgs: bool = False):
        return _fetch_batch_user_metadata_by_user_ids(
            self.auth_url, self.integration_api_key, user_ids, include_orgs
        )

    def fetch_batch_user_metadata_by_emails(self, emails: list[str], include_orgs: bool = False):
        return _fetch_batch_user_metadata_by_emails(
            self.auth_url, self.integration_api_key, emails, include_orgs
        )

    def fetch_batch_user_metadata_by_usernames(self, usernames: list[str], include_orgs: bool = False):
        return _fetch_batch_user_metadata_by_usernames(
            self.auth_url, self.integration_api_key, usernames, include_orgs
        )

    def fetch_org(self, org_id: str):
        return _fetch_org(self.auth_url, self.integration_api_key, org_id)

    def fetch_org_by_query(
        self, page_size: int = 10, page_number: int = 0, order_by: OrgQueryOrderBy = OrgQueryOrderBy.CREATED_AT_ASC, name: Optional[str] = None, legacy_org_id: Optional[str] = None, domain: Optional[str] = None
    ):
        return _fetch_org_by_query(
            self.auth_url,
            self.integration_api_key,
            page_size,
            page_number,
            order_by,
            name,
            legacy_org_id,
            domain,
        )

    def fetch_custom_role_mappings(self):
        return _fetch_custom_role_mappings(
            self.auth_url,
            self.integration_api_key,
        )

    def fetch_pending_invites(self, page_number: int = 0, page_size: int = 10, org_id: Optional[str] = None):
        return _fetch_pending_invites(
            self.auth_url,
            self.integration_api_key,
            page_number,
            page_size,
            org_id,
        )

    def fetch_users_by_query(
        self,
        page_size: int = 10,
        page_number: int = 0,
        order_by: UserQueryOrderBy = UserQueryOrderBy.CREATED_AT_ASC,
        email_or_username: Optional[str] = None,
        include_orgs: bool = False,
        legacy_user_id: Optional[str] = None
    ):
        return _fetch_users_by_query(
            self.auth_url,
            self.integration_api_key,
            page_size,
            page_number,
            order_by,
            email_or_username,
            include_orgs,
            legacy_user_id
        )

    def fetch_users_in_org(
        self, org_id: str, page_size: int = 10, page_number: int = 0, include_orgs: bool = False, role: Optional[str] = None
    ):
        return _fetch_users_in_org(
            self.auth_url,
            self.integration_api_key,
            org_id,
            page_size,
            page_number,
            include_orgs,
            role,
        )

    def create_user(
        self,
        email: str,
        email_confirmed: bool = False,
        send_email_to_confirm_email_address: bool = True,
        ask_user_to_update_password_on_login: bool = False,
        password: Optional[str] = None,
        username: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        properties: Optional[Dict[str, Any]] = None,
    ):
        return _create_user(
            self.auth_url,
            self.integration_api_key,
            email,
            email_confirmed,
            send_email_to_confirm_email_address,
            ask_user_to_update_password_on_login,
            password,
            username,
            first_name,
            last_name,
            properties,
        )

    def invite_user_to_org(self, email: str, org_id: str, role: str, additional_roles: list[str] = []):
        return _invite_user_to_org(
            self.auth_url,
            self.integration_api_key,
            email,
            org_id,
            role,
            additional_roles,
        )

    def resend_email_confirmation(self, user_id: str):
        return _resend_email_confirmation(
            self.auth_url,
            self.integration_api_key,
            user_id,
        )

    def logout_all_user_sessions(self, user_id: str):
        return _logout_all_user_sessions(
            self.auth_url,
            self.integration_api_key,
            user_id,
        )

    def update_user_email(self, user_id: str, new_email: str, require_email_confirmation: bool):
        return _update_user_email(
            self.auth_url,
            self.integration_api_key,
            user_id,
            new_email,
            require_email_confirmation,
        )

    def update_user_metadata(
        self,
        user_id: str,
        username: Optional[str] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        properties: Optional[Dict[str, Any]] = None,
        picture_url: Optional[str] = None,
        update_password_required: Optional[bool] = None,
    ):
        return _update_user_metadata(
            self.auth_url,
            self.integration_api_key,
            user_id=user_id,
            username=username,
            first_name=first_name,
            last_name=last_name,
            metadata=metadata,
            properties=properties,
            picture_url=picture_url,
            update_password_required=update_password_required,
        )

    def clear_user_password(self, user_id: str):
        return _clear_user_password(self.auth_url, self.integration_api_key, user_id)

    def update_user_password(
        self, user_id: str, password: str, ask_user_to_update_password_on_login: bool = False
    ):
        return _update_user_password(
            self.auth_url,
            self.integration_api_key,
            user_id,
            password,
            ask_user_to_update_password_on_login,
        )

    def create_magic_link(
        self, 
        email: str,
        redirect_to_url: Optional[str] = None,
        expires_in_hours: Optional[str] = None,
        create_new_user_if_one_doesnt_exist: Optional[bool] = None,
        user_signup_query_parameters: Optional[Dict[str, str]] = None,
    ):
        return _create_magic_link(
            self.auth_url,
            self.integration_api_key,
            email,
            redirect_to_url,
            expires_in_hours,
            create_new_user_if_one_doesnt_exist,
            user_signup_query_parameters,
        )

    def create_access_token(self, user_id: str, duration_in_minutes: int):
        return _create_access_token(
            self.auth_url, self.integration_api_key, user_id, duration_in_minutes
        )

    def migrate_user_from_external_source(
        self,
        email: str,
        email_confirmed: bool,
        existing_user_id: Optional[str] = None,
        existing_password_hash: Optional[str] = None,
        existing_mfa_base32_encoded_secret: Optional[str] = None,
        ask_user_to_update_password_on_login: bool = False,
        enabled: Optional[bool] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        username: Optional[str] = None,
        picture_url: Optional[str] = None,
        properties: Optional[Dict[str, Any]] = None,
    ):
        return _migrate_user_from_external_source(
            self.auth_url,
            self.integration_api_key,
            email,
            email_confirmed,
            existing_user_id,
            existing_password_hash,
            existing_mfa_base32_encoded_secret,
            ask_user_to_update_password_on_login,
            enabled,
            first_name,
            last_name,
            username,
            picture_url,
            properties,
        )

    def create_org(
        self,
        name: str,
        enable_auto_joining_by_domain: bool = False,
        members_must_have_matching_domain: bool = False,
        domain: Optional[str] = None,
        max_users: Optional[str] = None,
        custom_role_mapping_name: Optional[str] = None,
        legacy_org_id: Optional[str] = None,
    ):
        return _create_org(
            self.auth_url,
            self.integration_api_key,
            name,
            enable_auto_joining_by_domain,
            members_must_have_matching_domain,
            domain,
            max_users,
            custom_role_mapping_name,
            legacy_org_id,
        )

    def update_org_metadata(
        self,
        org_id: str,
        name: Optional[str] = None,
        can_setup_saml: Optional[bool] = None,
        metadata: Optional[Dict[str, Any]] = None,
        max_users: Optional[str] = None,
        can_join_on_email_domain_match: Optional[bool] = None,
        members_must_have_email_domain_match: Optional[bool] = None,
        domain: Optional[str] = None,
    ):
        return _update_org_metadata(
            self.auth_url,
            self.integration_api_key,
            org_id=org_id,
            name=name,
            can_setup_saml=can_setup_saml,
            metadata=metadata,
            max_users=max_users,
            can_join_on_email_domain_match=can_join_on_email_domain_match,
            members_must_have_email_domain_match=members_must_have_email_domain_match,
            domain=domain,
        )

    def subscribe_org_to_role_mapping(self, org_id: str, custom_role_mapping_name: str):
        return _subscribe_org_to_role_mapping(
            self.auth_url,
            self.integration_api_key,
            org_id,
            custom_role_mapping_name,
        )
        
    def fetch_saml_sp_metadata(self, org_id: str):
        return _fetch_saml_sp_metadata(
            self.auth_url,
            self.integration_api_key,
            org_id,
        )
        
    def set_saml_idp_metadata(self, org_id: str, saml_idp_metadata: SamlIdpMetadata):
        return _set_saml_idp_metadata(
            self.auth_url,
            self.integration_api_key,
            org_id,
            saml_idp_metadata,
        )
        
    def saml_go_live(self, org_id: str):
        return _saml_go_live(
            self.auth_url,
            self.integration_api_key,
            org_id,
        )
        
    def delete_saml_connection(self, org_id: str):
        return _delete_saml_connection(
            self.auth_url,
            self.integration_api_key,
            org_id,
        )

    def delete_org(self, org_id: str):
        return _delete_org(self.auth_url, self.integration_api_key, org_id)
    
    def revoke_pending_org_invite(self, org_id: str, invitee_email: str):
        return _revoke_pending_org_invite(self.auth_url, self.integration_api_key, org_id, invitee_email)


    def add_user_to_org(self, user_id: str, org_id: str, role: str, additional_roles: list[str] = []):
        return _add_user_to_org(
            self.auth_url, self.integration_api_key, user_id, org_id, role, additional_roles
        )

    def remove_user_from_org(self, user_id: str, org_id: str):
        return _remove_user_from_org(self.auth_url, self.integration_api_key, user_id, org_id)

    def change_user_role_in_org(self, user_id: str, org_id: str, role: str, additional_roles: list[str] = []):
        return _change_user_role_in_org(
            self.auth_url, self.integration_api_key, user_id, org_id, role, additional_roles
        )

    def delete_user(self, user_id: str):
        return _delete_user(self.auth_url, self.integration_api_key, user_id)

    def disable_user(self, user_id: str):
        return _disable_user(self.auth_url, self.integration_api_key, user_id)

    def enable_user(self, user_id: str):
        return _enable_user(self.auth_url, self.integration_api_key, user_id)

    def disable_user_2fa(self, user_id: str):
        return _disable_user_2fa(self.auth_url, self.integration_api_key, user_id)

    def enable_user_can_create_orgs(self, user_id: str):
        return _enable_user_can_create_orgs(self.auth_url, self.integration_api_key, user_id)

    def disable_user_can_create_orgs(self, user_id: str):
        return _disable_user_can_create_orgs(self.auth_url, self.integration_api_key, user_id)

    def allow_org_to_setup_saml_connection(self, org_id: str):
        return _allow_org_to_setup_saml_connection(
            self.auth_url, self.integration_api_key, org_id
        )

    def disallow_org_to_setup_saml_connection(self, org_id: str):
        return _disallow_org_to_setup_saml_connection(
            self.auth_url, self.integration_api_key, org_id
        )

    def create_org_saml_connection_link(self, org_id: str, expires_in_seconds=None):
        return _create_org_saml_connection_link(
            self.auth_url, self.integration_api_key, org_id, expires_in_seconds
        )

    # functions for end user api keys

    def fetch_api_key(self, api_key_id: str):
        return _fetch_api_key(self.auth_url, self.integration_api_key, api_key_id)

    def fetch_current_api_keys(
        self,
        org_id: Optional[str] = None,
        user_id: Optional[str] = None,
        user_email: Optional[str] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        api_key_type: Optional[str] = None,
    ):
        return _fetch_current_api_keys(
            self.auth_url,
            self.integration_api_key,
            org_id,
            user_id,
            user_email,
            page_size,
            page_number,
            api_key_type,
        )

    def fetch_archived_api_keys(
        self,
        org_id: Optional[str] = None,
        user_id: Optional[str] = None,
        user_email: Optional[str] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        api_key_type: Optional[str] = None,
    ):
        return _fetch_archived_api_keys(
            self.auth_url,
            self.integration_api_key,
            org_id,
            user_id,
            user_email,
            page_size,
            page_number,
            api_key_type=api_key_type,
        )

    def create_api_key(
        self, org_id: Optional[str] = None, user_id: Optional[str] = None, expires_at_seconds: Optional[str] = None, metadata: Optional[Dict[str, Any]] = None
    ):
        return _create_api_key(
            self.auth_url, self.integration_api_key, org_id, user_id, expires_at_seconds, metadata
        )

    def update_api_key(self, api_key_id: str, expires_at_seconds: Optional[str] = None, metadata: Optional[Dict[str, Any]] = None):
        return _update_api_key(
            self.auth_url, self.integration_api_key, api_key_id, expires_at_seconds, metadata
        )

    def delete_api_key(self, api_key_id: str):
        return _delete_api_key(self.auth_url, self.integration_api_key, api_key_id)

    def validate_personal_api_key(self, api_key_token: str):
        return _validate_personal_api_key(self.auth_url, self.integration_api_key, api_key_token)

    def validate_org_api_key(self, api_key_token: str):
        return _validate_org_api_key(self.auth_url, self.integration_api_key, api_key_token)

    def validate_api_key(self, api_key_token: str):
        return _validate_api_key(self.auth_url, self.integration_api_key, api_key_token)
    

    def validate_access_token_and_get_user(self, authorization_header: Optional[str]):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, self.token_verification_metadata)
        return user
    
    def validate_access_token_and_get_user_with_org(self, authorization_header: Optional[str], required_org_id: str):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, self.token_verification_metadata)
        org_member_info = validate_org_access_and_get_org_member_info(user, required_org_id)
        return UserAndOrgMemberInfo(user, org_member_info)
    
    def validate_access_token_and_get_user_with_org_by_minimum_role(self, authorization_header: Optional[str], required_org_id: str, minimum_required_role: str):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, self.token_verification_metadata)
        org_member_info = validate_minimum_org_role_and_get_org(user, required_org_id, minimum_required_role)
        return UserAndOrgMemberInfo(user, org_member_info)
    
    def validate_access_token_and_get_user_with_org_by_exact_role(self, authorization_header: Optional[str], required_org_id: str, required_role: str):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, self.token_verification_metadata)
        org_member_info = validate_exact_org_role_and_get_org(user, required_org_id, required_role)
        return UserAndOrgMemberInfo(user, org_member_info)
    
    def validate_access_token_and_get_user_with_org_by_permission(self, authorization_header: Optional[str], required_org_id: str, permission: str):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, self.token_verification_metadata)
        org_member_info = validate_permission_and_get_org(user, required_org_id, permission)
        return UserAndOrgMemberInfo(user, org_member_info)
    
    def validate_access_token_and_get_user_with_org_by_all_permissions(self, authorization_header: Optional[str], required_org_id: str, permissions: list[str]):
        access_token = _extract_token_from_authorization_header(authorization_header)
        user = _validate_access_token_and_get_user(access_token, self.token_verification_metadata)
        org_member_info = validate_all_permissions_and_get_org(user, required_org_id, permissions)
        return UserAndOrgMemberInfo(user, org_member_info)
    
    def validate_org_access_and_get_org(self, user: User, required_org_id: str):
        return validate_org_access_and_get_org_member_info(user, required_org_id)
    
    def validate_minimum_org_role_and_get_org(self, user: User, required_org_id: str, minimum_role: str):
        return validate_minimum_org_role_and_get_org(user, required_org_id, minimum_role)
    
    def validate_exact_org_role_and_get_org(self, user: User, required_org_id: str, exact_role: str):
        return validate_exact_org_role_and_get_org(user, required_org_id, exact_role)
    
    def validate_permission_and_get_org(self, user: User, required_org_id: str, permission: str):
        return validate_permission_and_get_org(user, required_org_id, permission)
    
    def validate_all_permissions_and_get_org(self, user: User, required_org_id: str, permissions: list[str]):
        return validate_all_permissions_and_get_org(user, required_org_id, permissions)
    


def init_base_auth(
    auth_url: str,
    integration_api_key: str,
    token_verification_metadata: Optional[TokenVerificationMetadata] = None,
) -> Auth:
    auth_url = _validate_url(auth_url)
    token_verification_metadata = _fetch_token_verification_metadata(
        auth_url, integration_api_key, token_verification_metadata
    )
    return Auth(auth_url, integration_api_key, token_verification_metadata)


from urllib.parse import urlparse


def _validate_url(auth_url):
    parsed_url = urlparse(auth_url)
    if parsed_url.scheme != "https":
        raise ValueError("URL must start with https://")
    elif parsed_url.netloc == "":
        raise ValueError("Invalid URL")
    else:
        return "https://" + parsed_url.netloc



from typing import Optional
import requests

from propelauth_py.api import _ApiKeyAuth, _format_params, _is_valid_id
from propelauth_py.api.end_user_api_keys import _validate_api_key
from propelauth_py.types.user import Organization, OrgQueryResponse, Org, PendingInvite, PendingInvitesPage, CreatedOrg, OrgApiKeyValidation
from propelauth_py.types.custom_role_mappings import CustomRoleMappings, CustomRoleMapping
from propelauth_py.types.saml_types import SamlIdpMetadata, SpMetadata
from propelauth_py.errors import (
    BadRequestException,
    EndUserApiKeyException,
    UpdateUserMetadataException,
)

BASE_ENDPOINT_PATH = "/api/backend/v1"
ORG_ENDPOINT_PATH = "/api/backend/v1/org"


####################
#       GET        #
####################
def _fetch_org(auth_url, integration_api_key, org_id) -> Optional[Organization]:
    if not _is_valid_id(org_id):
        return None

    url = auth_url + f"{ORG_ENDPOINT_PATH}/{org_id}"
    response = requests.get(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return None
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching org")

    json_response = response.json()
    return Organization(
        org_id=json_response.get('org_id'),
        name=json_response.get('name'),
        url_safe_org_slug=json_response.get('url_safe_org_slug'),
        can_setup_saml=json_response.get('can_setup_saml'),
        is_saml_configured=json_response.get('is_saml_configured'),
        is_saml_in_test_mode=json_response.get('is_saml_in_test_mode'),
        max_users=json_response.get('max_users'),
        metadata=json_response.get('metadata'),
        domain=json_response.get('domain'),
        domain_autojoin=json_response.get('domain_autojoin'),
        domain_restrict=json_response.get('domain_restrict'),
        custom_role_mapping_name=json_response.get('custom_role_mapping_name'),
        legacy_org_id=json_response.get('legacy_org_id')
    )


def _fetch_org_by_query(
    auth_url, integration_api_key, page_size, page_number, order_by, name, legacy_org_id, domain
) -> OrgQueryResponse:
    url = auth_url + f"{ORG_ENDPOINT_PATH}/query"
    params = {
        "page_size": page_size,
        "page_number": page_number,
        "order_by": order_by,
        "name": name,
        "legacy_org_id": legacy_org_id,
        "domain": domain,
    }
    response = requests.get(
        url, params=_format_params(params), auth=_ApiKeyAuth(integration_api_key)
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise ValueError("Bad request: " + response.text)
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching orgs by query")

    json_response = response.json()
    
    orgs = [
        Org(
            org_id=key.get('org_id'),
            name=key.get('name'),
            max_users=key.get('max_users'),
            is_saml_configured=key.get('is_saml_configured'),
            legacy_org_id=key.get('legacy_org_id'),
            metadata=key.get('metadata'),
            custom_role_mapping_name=key.get('custom_role_mapping_name')
        )
        for key in json_response.get('orgs')
    ]
    
    return OrgQueryResponse(
        orgs=orgs,
        total_orgs=json_response.get('total_orgs'),
        current_page=json_response.get('current_page'),
        page_size=json_response.get('page_size'),
        has_more_results=json_response.get('has_more_results')
    )


def _fetch_custom_role_mappings(auth_url, integration_api_key) -> CustomRoleMappings:
    url = auth_url + "/api/backend/v1/custom_role_mappings"
    response = requests.get(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching org")

    json_response = response.json()
    
    role_mappings = [
        CustomRoleMapping(
            custom_role_mapping_name=key.get('custom_role_mapping_name'),
            num_orgs_subscribed=key.get('num_orgs_subscribed')
        )
        for key in json_response.get('custom_role_mappings')
    ]
    
    return CustomRoleMappings(
        custom_role_mappings=role_mappings
    )


def _fetch_pending_invites(
    auth_url,
    integration_api_key,
    page_number=0,
    page_size=10,
    org_id=None,
) -> Optional[PendingInvitesPage]:
    if org_id:
        if not _is_valid_id(org_id):
            return None

    url = auth_url + "/api/backend/v1/pending_org_invites"
    params = {
        "page_number": page_number,
        "page_size": page_size,
        "org_id": org_id,
    }

    response = requests.get(
        url, params=_format_params(params), auth=_ApiKeyAuth(integration_api_key)
    )
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching pending invites")

    json_response = response.json()
    
    invites = [
        PendingInvite(
            invitee_email=key.get('invitee_email'),
            org_id=key.get('org_id'),
            org_name=key.get('org_name'),
            role_in_org=key.get('role_in_org'),
            additional_roles_in_org=key.get('additional_roles_in_org'),
            created_at=key.get('created_at'),
            expires_at=key.get('expires_at'),
            inviter_email=key.get('inviter_email'),
            inviter_user_id=key.get('inviter_user_id'),
        )
        for key in json_response.get('invites')
    ]
    
    return PendingInvitesPage(
        invites=invites,
        total_invites=json_response.get('total_invites'),
        current_page=json_response.get('current_page'),
        page_size=json_response.get('page_size'),
        has_more_results=json_response.get('has_more_results')
    )
    
def _fetch_saml_sp_metadata(auth_url, integration_api_key, org_id) -> Optional[SpMetadata]:
    if not _is_valid_id(org_id):
        return None

    url = auth_url + f"{BASE_ENDPOINT_PATH}/saml_sp_metadata/{org_id}"
    response = requests.get(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return None
    elif not response.ok:
        raise RuntimeError("Unknown error when fetching org SAML SP metadata")

    json_response = response.json()
    return SpMetadata(
        entity_id=json_response.get('entity_id'),
        acs_url=json_response.get('acs_url'),
        logout_url=json_response.get('logout_url'),
    )


####################
#       POST       #
####################
def _create_org(
    auth_url,
    integration_api_key,
    name,
    enable_auto_joining_by_domain=False,
    members_must_have_matching_domain=False,
    domain=None,
    max_users=None,
    custom_role_mapping_name=None,
    legacy_org_id=None,
) -> CreatedOrg:
    url = auth_url + f"{ORG_ENDPOINT_PATH}/"
    json = {
        "name": name,
        "enable_auto_joining_by_domain": enable_auto_joining_by_domain,
        "members_must_have_matching_domain": members_must_have_matching_domain,
    }
    if domain:
        json["domain"] = domain
    if max_users is not None:
        json["max_users"] = max_users
    if legacy_org_id:
        json["legacy_org_id"] = legacy_org_id
    if custom_role_mapping_name is not None:
        json["custom_role_mapping_name"] = custom_role_mapping_name

    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when creating an org")

    json_response = response.json()
    return CreatedOrg(
        org_id=json_response.get('org_id'),
        name=json_response.get('name')
    )


def _allow_org_to_setup_saml_connection(auth_url, integration_api_key, org_id) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = auth_url + f"{ORG_ENDPOINT_PATH}/{org_id}/allow_saml"
    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when allowing org to setup SAML connection")

    return True


def _disallow_org_to_setup_saml_connection(auth_url, integration_api_key, org_id) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = auth_url + f"{ORG_ENDPOINT_PATH}/{org_id}/disallow_saml"
    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when allowing org to setup SAML connection")

    return True


def _create_org_saml_connection_link(
    auth_url, integration_api_key, org_id, expires_in_seconds=None
):
    if not _is_valid_id(org_id):
        return None

    url = auth_url + f"{ORG_ENDPOINT_PATH}/{org_id}/create_saml_connection_link"

    body = {}
    if expires_in_seconds is not None:
        body["expires_in_seconds"] = expires_in_seconds

    response = requests.post(url, json=body, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 426:
        raise RuntimeError(
            "Cannot use organizations unless B2B support is enabled. Enable it in your PropelAuth "
            "dashboard."
        )
    elif not response.ok:
        raise RuntimeError("Unknown error when creating org SAML connection link")

    return response.json()


def _add_user_to_org(
    auth_url, integration_api_key, user_id, org_id, role, additional_roles=[]
) -> bool:
    url = auth_url + f"{ORG_ENDPOINT_PATH}/add_user"
    json = {
        "user_id": user_id,
        "org_id": org_id,
        "role": role,
        "additional_roles": additional_roles,
    }

    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when adding a user to the org")

    return True


def _remove_user_from_org(auth_url, integration_api_key, user_id, org_id) -> bool:
    url = auth_url + f"{ORG_ENDPOINT_PATH}/remove_user"
    json = {"user_id": user_id, "org_id": org_id}

    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when removing a user from the org")

    return True


def _change_user_role_in_org(
    auth_url, integration_api_key, user_id, org_id, role, additional_roles=[]
) -> bool:
    url = auth_url + f"{ORG_ENDPOINT_PATH}/change_role"
    json = {
        "user_id": user_id,
        "org_id": org_id,
        "role": role,
        "additional_roles": additional_roles,
    }

    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when changing a user's role(s) in the org")

    return True

def _set_saml_idp_metadata(auth_url, integration_api_key, org_id, saml_idp_metadata: SamlIdpMetadata) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = auth_url + f"{BASE_ENDPOINT_PATH}/saml_idp_metadata"

    required_fields = ["idp_entity_id", "idp_sso_url", "idp_certificate", "provider"]
    json = {"org_id": org_id}
    
    for field in required_fields:
        if field not in saml_idp_metadata:
            raise ValueError(f"Missing required field '{field}' in SAML IdP metadata")
        json[field] = saml_idp_metadata[field]

    response = requests.post(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when setting SAML IdP metadata")

    return True

def _saml_go_live(auth_url, integration_api_key, org_id) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = auth_url + f"{BASE_ENDPOINT_PATH}/saml_idp_metadata/go_live/{org_id}"

    response = requests.post(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when setting SAML connection to go live")

    return True

####################
#     PATCH/PUT    #
####################
def _update_org_metadata(
    auth_url,
    integration_api_key,
    org_id,
    name=None,
    can_setup_saml=None,
    metadata=None,
    max_users=None,
    can_join_on_email_domain_match=None,  # In the backend, this is the `domain_autojoin` argument.
    members_must_have_email_domain_match=None,  # In the backend, this is the `domain_restrict` argument.
    domain=None,
    legacy_org_id=None,
    # TODO: Add `require_2fa_by` optional argument.
) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = auth_url + f"{ORG_ENDPOINT_PATH}/{org_id}"
    json = {}
    if name is not None:
        json["name"] = name
    if can_setup_saml is not None:
        json["can_setup_saml"] = can_setup_saml
    if metadata is not None:
        json["metadata"] = metadata
    if max_users is not None:
        json["max_users"] = max_users
    if can_join_on_email_domain_match is not None:
        json["autojoin_by_domain"] = can_join_on_email_domain_match
    if members_must_have_email_domain_match is not None:
        json["restrict_to_domain"] = members_must_have_email_domain_match
    if domain is not None:
        json["domain"] = domain
    if legacy_org_id is not None:
        json["legacy_org_id"] = legacy_org_id

    response = requests.put(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise UpdateUserMetadataException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when updating org metadata")

    return True


def _subscribe_org_to_role_mapping(
    auth_url,
    integration_api_key,
    org_id,
    custom_role_mapping_name,
) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = auth_url + f"{ORG_ENDPOINT_PATH}/{org_id}"
    json = {
        "custom_role_mapping_name": custom_role_mapping_name,
    }

    response = requests.put(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise UpdateUserMetadataException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError(
            "Unknown error when subscribing an org to a custom role mapping"
        )

    return True


####################
#      DELETE      #
####################


def _delete_org(auth_url, integration_api_key, org_id) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = auth_url + f"{ORG_ENDPOINT_PATH}/{org_id}"
    response = requests.delete(url, auth=_ApiKeyAuth(integration_api_key))

    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when deleting org")

    return True

def _revoke_pending_org_invite(auth_url, integration_api_key, org_id, invitee_email) -> bool:

    url = auth_url + "/api/backend/v1/pending_org_invites"
    json = {"org_id": org_id, "invitee_email": invitee_email}

    response = requests.delete(url, json=json, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif not response.ok:
        raise RuntimeError("Unknown error when revoking pending org invite")

    return response.json()


def _delete_saml_connection(auth_url, integration_api_key, org_id) -> bool:
    if not _is_valid_id(org_id):
        return False

    url = auth_url + f"{BASE_ENDPOINT_PATH}/saml_idp_metadata/{org_id}"

    response = requests.delete(url, auth=_ApiKeyAuth(integration_api_key))
    if response.status_code == 401:
        raise ValueError("integration_api_key is incorrect")
    elif response.status_code == 400:
        raise BadRequestException(response.json())
    elif response.status_code == 404:
        return False
    elif not response.ok:
        raise RuntimeError("Unknown error when deleting SAML connection")

    return True


####################
#      HELPERS     #
####################


def _validate_org_api_key(auth_url, integration_api_key, api_key_token) -> OrgApiKeyValidation:
    api_key_validation = _validate_api_key(auth_url, integration_api_key, api_key_token)
    if not api_key_validation.org:
        raise EndUserApiKeyException({"api_key_token": ["Not an org API Key"]})
    return OrgApiKeyValidation(
        org=api_key_validation.org,
        metadata=api_key_validation.metadata,
        user=api_key_validation.user,
        user_in_org=api_key_validation.user_in_org,
    )
